package cn.ischoolcode.javabook.dushichat;

/* 
 * 爱校码
 * http://www.ischoolcode.cn/
 *  
 */

import java.net.*;
import java.util.*;
import java.text.*;
import java.io.*;

/**
 * 此类表示在服务器端用户的连接。所有的通信
 * 由服务器通过该类进行。
 * 
 * @zcj
 */
public class DushiServerConnection extends DushiConnection {
	private Class thisClass = DushiServerConnection.class;
	private Date connectDate;
	protected DushiServer server;
	protected DushiUser user;
	protected DushiStringManager strings = null;

	public DushiServerConnection(ThreadGroup tg, Socket s, DushiServer parent)
			throws IOException {
		super(tg, "Dushi Chat server connection", s);

		server = parent;

		// 获取“用户”对象
		user = new DushiUser(server);

		// 初始化，使用服务器的语言。
		strings = server.strings;
		user.language = strings.language;

		// 建立一个初始的虚拟聊天室。
		user.chatroom = new DushiChatRoom("", "", false, "");

		// 记录连接的日期/时间。
		connectDate = new Date();

		start();

		// 最后，启动一个看门狗来监控连接。
		if (protocolVersion >= 2.2)
			watchDog = new WatchDog(this);
	}

	public void run() {
		try {
			// 等待客户端开始协议协商。
			synchronized (iStream) {
				short command = iStream.readShort();

				if (command != DushiCommand.SETPROTO) {
					server.serverOutput(strings.get(thisClass, "noprotocol")
							+ "\n");
					server.disconnect(this, false);
					return;
				}

				receiveProtocol();
			}
		} catch (IOException e) {
			// 这并不顺利。甚至无法启动连接。
			server.disconnect(this, false);
			return;
		}

		while (online) {
			try {
				parseCommand();
			} catch (EOFException e) {
				// 这种情况偶尔会发生，就好像Java流认为某个东西在实际读取
				// 之前就可以立即读取一样……只需重置到标记的位置。
				if (online) {
					try {
						iStream.reset();
					} catch (IOException f) {
						lostConnection();
					}
				}
			} catch (IOException e) {
				lostConnection();
			} catch (Exception e) {
				// 这里发现的任何其他异常都肯定是某种错误。并将其记录在日志中。
				// 将异常堆栈跟踪转换为字符串。
				StringWriter sw = new StringWriter();
				e.printStackTrace(new PrintWriter(sw));

				server.serverOutput(strings.get(thisClass, "problem") + ": \n"
						+ sw.toString());

				// 如果可能的话，尽量忽略它。
			}

			yield();
		}
	}

	private void parseCommand() throws IOException {
		short commandType = 0;
		boolean foundCommand = false;

		synchronized (iStream) {

			// 我们需要在这个“同步”块中进行所有输入处理，因为我们在这里读取命令
			// 的第一部分，然后在子例程中读取其余部分。

			// 标记当前点，以便在必要时返回到该点。
			iStream.mark(Integer.MAX_VALUE);

			// 试着从流里读一篇短文。它指示命令类型。
			commandType = iStream.readShort();

			// 记录我们收到了什么。
			lastReceiveTime = System.currentTimeMillis();

			// 请确保在阻塞输入时没有断开连接。
			if (!online)
				return;

			// 现在，我们检查正在传递的消息，并确定如何处理它。

			if (!foundCommand && (protocolVersion >= 2.2)) {
				// 协议版本2.2支持以下命令。
				// 有关协议定义，请参阅文件DushiCommand.java。

				foundCommand = true;

				switch (commandType) {

				case DushiCommand.SENDEMAIL: {
					// 我们收到发送电子邮件的请求。
					receiveSendEmail();
					break;
				}
				case DushiCommand.FILE: {
					// 我们收到启动文件传输的请求。
					receiveFile();
					break;
				}
				case DushiCommand.FILEACCEPT: {
					//我们收到关于用户是否已接受文件传输的通知。
					receiveFileAccept();
					break;
				}
				case DushiCommand.FILEBLOCK: {
					// 我们收到一个文件块，它是文件传输的一部分。
					receiveFileBlock();
					break;
				}
				case DushiCommand.FILECANCEL: {
					// 我们收到通知，文件传输已被取消。
					receiveFileCancel();
					break;
				}
				case DushiCommand.IMAGEBLOCK: {
					// 我们正在接收一个图像块，它是图像传输的一部分。
					receiveImageBlock();
					break;
				}
				default: {
					foundCommand = false;
					break;
				}
				}
			}

			if (!foundCommand && (protocolVersion >= 2.0)) {
				// 协议版本2.0支持以下命令。

				foundCommand = true;

				switch (commandType) {

				case DushiCommand.NOOP: {
					// 不执行任何操作
					break;
				}
				case DushiCommand.PING: {
					// 我们收到客户端的ping指令。
					break;
				}
				case DushiCommand.USERINFO: {
					// 新用户正在提供有关他们自己的信息。
					receiveUserInfo();
					break;
				}
				case DushiCommand.DISCONNECT: {
					// 用户告诉我们他正在断开连接。
					receiveDisconnect();
					break;
				}
				case DushiCommand.ROOMLIST: {
					// 客户端正在请求聊天室的列表。
					receiveRoomList();
					break;
				}
				case DushiCommand.INVITE: {
					// 用户正在邀请其他用户加入聊天室。
					receiveInvite();
					break;
				}
				case DushiCommand.ENTERROOM: {
					// 客户端正在尝试“创建/进入”聊天室。
					receiveEnterRoom();
					break;
				}
				case DushiCommand.BOOTUSER: {
					// 客户端正试图从聊天室启动另一个用户。
					receiveBootUser();
					break;
				}
				case DushiCommand.BANUSER: {
					// 客户端正试图禁止其他用户进入聊天室。
					receiveBanUser();
					break;
				}
				case DushiCommand.ALLOWUSER: {
					// 客户端希望允许被禁止的用户返回聊天室。
					receiveAllowUser();
					break;
				}
				case DushiCommand.ACTIVITY: {
					// 客户端正在发送有关用户活动的消息。
					receiveActivity();
					break;
				}
				case DushiCommand.CHATTEXT: {
					// 客户端正在发送一行聊天文本。
					receiveChatText();
					break;
				}
				case DushiCommand.LINE: {
					// 用户在画布上画了一条线。
					receiveLine();
					break;
				}
				case DushiCommand.RECT: {
					// 用户在画布上画了一个矩形。
					receivePoly(DushiCommand.RECT);
					break;
				}
				case DushiCommand.OVAL: {
					// 用户在画布上画了一个椭圆形。
					receivePoly(DushiCommand.OVAL);
					break;
				}
				case DushiCommand.DRAWTEXT: {
					// 用户在画布上画了一些文字。
					receiveDrawText();
					break;
				}
				case DushiCommand.IMAGE: {
					// 用户在画布上粘贴了一个图像。
					receiveImage();
					break;
				}
				case DushiCommand.CLEARCANV: {
					// 此客户端正在清理画布。
					receiveClearCanv();
					break;
				}
				case DushiCommand.PAGEUSER: {
					// 用户正在寻呼其他客户端
					receivePageUser();
					break;
				}
				case DushiCommand.INSTANTMESS: {
					// 客户端想要向某人发送即时消息
					receiveInstantMess();
					break;
				}
				case DushiCommand.LEAVEMESS: {
					// 客户想给某人留言
					receiveLeaveMess();
					break;
				}
				case DushiCommand.READMESS: {
					// 客户端希望收集其消息
					receiveReadMess();
					break;
				}
				case DushiCommand.ERROR: {
					// 客户端正在发送错误信息
					receiveError();
					break;
				}
				default: {
					foundCommand = false;
					break;
				}
				}
			}

			if (!foundCommand) {
				// 哎呀。我们不理解这个命令，

				byte[] foo;

				//获取流中的任何数据。
				foo = new byte[iStream.available()];
				iStream.readFully(foo);

				// 嗯。这个客户似乎在说一种我们不懂的语言。
				// 它可能是Dushi Chat 1.x客户端吗？
				// 让我们检查一下里面是否有1.x时代的东西。
				String stringFoo = new String(foo);
				if (stringFoo.startsWith("AME ")) {
					// 这可能是一个1.x客户端。
					// 向客户端发送1.x样式的错误消息，
					// 然后断开它们的连接。
					notifyObsoleteV1();
					shutdown();
					return;
				}

				server.serverOutput(strings.get(thisClass, "unknown") + " ("
						+ user.name + ")");
			}
		}
	}

	private void notifyObsoleteV1() throws IOException {
		// 此功能用于通知客户端此服务器无法接受Dushi Chat v1.x客户端的连接。
		
		PrintWriter tmpOutput = new PrintWriter(socket.getOutputStream(), true);

		tmpOutput.println("/SERVERMESSAGE "
				+ strings.get(thisClass, "incompatible"));
		tmpOutput.println("/SERVERMESSAGE "
				+ strings.get(thisClass, "disconnected"));
		tmpOutput.flush();
	}

	private DushiServerConnection findConnection(int userid) {
		// 这将为具有请求ID的用户返回连接对象。

		DushiServerConnection tempConnection = null;
		DushiServerConnection returnConnection = null;

		synchronized (server.connections) {
			for (int count = 0; count < server.currentConnections; count++) {
				tempConnection = (DushiServerConnection) server.connections
						.elementAt(count);

				if ((tempConnection.user.id == userid)
						&& (tempConnection != this)) {
					returnConnection = tempConnection;
					break;
				}
			}
		}

		return (returnConnection);
	}

	private void catchUpWithCanvas(DushiChatRoom room) {
		for (int count = 0; count < user.chatroom.canvasHistory.size(); count++) {
			CanvasCommand command;
			try {
				command = (CanvasCommand) user.chatroom.canvasHistory
						.elementAt(count);
			} catch (ArrayIndexOutOfBoundsException e) {
				//以防有人在我们试图从向量中获取命令对象时清除画布。
				break;
			}

			// 发送每个绘图元素，具体取决于它是什么。
			switch (command.command) {
			  case DushiCommand.LINE: {
				  sendLine(command.who, command.colour, command.x1, command.y1,
						command.x2, command.y2, command.thick);
				  break;
			  }
			  case DushiCommand.RECT:
			  case DushiCommand.OVAL: {
				  sendPoly(command.command, command.who, command.colour,
						command.x1, command.y1, command.x2, command.y2,
						command.thick, command.fill);
				  break;
			  }
			  case DushiCommand.DRAWTEXT: {
				  sendDrawText(command.who, command.colour, command.x1,
						command.y1, command.type, command.attribs,
						command.size, command.text);
				  break;
			  }
			  case DushiCommand.IMAGE: {
				  if (protocolVersion >= 2.2) {
					int totalSize = command.imageData.length;
			  		int blockSize = DushiImageTransfer.DEFAULT_BLOCKSIZE;
					int totalBlocks = ((totalSize / blockSize) + (((totalSize % blockSize) > 0) ? 1
							: 0));
					Vector v = new Vector();
					v.add(this);
					DushiImageTransfer transfer = new DushiImageTransfer(
							command.who, command.imageName, command.x1,
							command.y1, totalSize, totalBlocks, blockSize, v,
							new int[0], null, null);
					transfer.putImageAll(command.imageData);
				  } else
					sendImage(command.who, null, command.x1, command.y1, 0, 0,
							0, command.imageData, new int[0]);
				  break;
			  }
			}
		}
	}

	private void enterChatRoom(DushiChatRoom room) {
		// 此例程将向聊天室添加一个用户，并向所有其他参与者宣布该用户。
		
		DushiServerConnection tmpConnection;

		// 将用户添加到聊天室。
		room.addUser(this);

		// 设置我们的聊天室指针
		user.chatroom = room;

		// 向包括此客户端在内的所有客户端发送“输入”消息
		synchronized (server.connections) {
			for (int count = 0; count < server.currentConnections; count++) {
				tmpConnection = (DushiServerConnection) server.connections
						.elementAt(count);

				tmpConnection.sendEnterRoom(user.id, user.chatroom);
			}
		}

		// 现在，让这个客户端了解聊天室中当前绘图画布上的最新内容。
		catchUpWithCanvas(room);

		// 完成。
		return;
	}

	/**
	 * 将该用户从聊天室中删除，并在需要时将其告知其他用户。
	 * 如果聊天室不是“主”聊天室，并且没有剩余的参与者，
	 * 它也会关闭聊天室。
	 */
	protected void leaveChatRoom() {
		DushiServerConnection tmpConnection;

		// 无论如何，如果没有聊天室，请退出。
		if (user.chatroom == null)
			return;

		// 将用户从聊天室中删除
		user.chatroom.removeUser(this);

		// 如果我们要离开主聊天室以外的某个聊天室，
		// 并且该聊天室中没有其他用户，我们将删除该聊天室。
		if (user.chatroom != server.mainRoom) {
			if (user.chatroom.users.size() == 0) {
				// 删除聊天室。
				server.chatRooms.removeElement(user.chatroom);
				server.chatRooms.trimToSize();

				synchronized (server.connections) {
					// 将更新后的聊天室列表发送给所有客户端。
					for (int count = 0; count < server.currentConnections; count++) {
						tmpConnection = (DushiServerConnection) server.connections
								.elementAt(count);
						tmpConnection.sendRoomList();
					}
				}
			}
		}

		// 完成。
		return;
	}

	// 以下例程将发送和接收各种命令类型。
	private void receiveProtocol() throws IOException {
		double preferred = 0;

		// 从输入流中读取客户端的首选协议版本。
		preferred = iStream.readDouble();

		// 向用户发送一条关于我们将使用哪种协议的消息。
		// 它将始终是所请求的或我们所知道的两者中较小的一个。

		if (preferred < protocolVersion)
			protocolVersion = preferred;

		sendProtocol(protocolVersion);
	}

	private void sendUserInfo(DushiUser newuser) {
		// 向此参与者发送有关新用户的信息。
		super.sendUserInfo(newuser.id, newuser.name, "", false, newuser.email,
				newuser.additional, newuser.language);
	}

	private void receiveUserInfo() throws IOException {
		DushiServerConnection tmpConnection;
		SimpleDateFormat dateFormatter = new SimpleDateFormat(
				"MMM dd, yyyy hh:mm a");

		// 接收有关用户的信息。
		iStream.readInt(); // 我们分配用户ID。
		user.name = iStream.readUTF();
		user.password = iStream.readUTF();
		if (!iStream.readBoolean())
			user.password = server.passwordEncryptor
					.encryptPassword(user.password);

		if (protocolVersion >= 2.2)
			// 读取用户的电子邮件地址。
			user.email = iStream.readUTF();

		user.additional = iStream.readUTF();

		if (protocolVersion >= 2.1)
			// 阅读用户的首选语言。
			user.language = iStream.readUTF();

		// 设置适合此用户的语言。
		strings = new DushiStringManager(server.baseURL, user.language);

		synchronized (server.connections) {
			// 请确保我们还没有同名的用户。
			// 循环通过所有当前连接的用户。
			for (int count = 0; count < server.currentConnections; count++) {
				tmpConnection = ((DushiServerConnection) server.connections
						.elementAt(count));

				if (tmpConnection.user.name.equals(user.name)) {
					sendDisconnect(0,
							strings.get(thisClass, "already") + " " + user.name
									+ " " + strings.get(thisClass, "loggedin"));

					// 在我们断开客户的连接之前，给他一点时间接收消息。
					try {
						sleep(1000);
					} catch (InterruptedException e) {
					}

					shutdown();
					return;
				}
			}
		}

		// 检查保留的用户名“Administrator”。
		// 此名称只能在与服务器从同一系统连接时使用（例如通过管理员控制台）。
		if (user.name.equals("Administrator")) {
			// 这是本地连接吗？
			InetAddress tmpAddr = socket.getInetAddress();
			if (!tmpAddr.getHostAddress().equals(
					InetAddress.getLocalHost().getHostAddress())
					&& !tmpAddr.getHostAddress().equals("127.0.0.1")) {
				sendDisconnect(0, strings.get(thisClass, "noadministrator"));
				server.serverOutput(tmpAddr.getHostName() + " "
						+ strings.get(thisClass, "badadminattempt") + "\n");

				// 在我们断开客户的连接之前，给他一点时间来接收消息。
				try {
					sleep(1000);
				} catch (InterruptedException e) {
				}

				shutdown();
				return;
			}

			// 在服务器中保存对此客户端套接字的引用，
			// 以便服务器可以轻松识别此连接的特殊状态。
			server.adminConnection = this;
		}

		// 如果我们需要密码，请确保密码与密码文件中的密码匹配。
		else if (server.requirePasswords) {
			boolean authenticated = false;

			try {
				authenticated = server.checkPassword(
						DushiServer.passwordFileName, user.name, user.password);
			} catch (Exception e) {
				// 此用户不存在。我们应该创造它吗？
				if (server.allowNewUsers) {
					// 创建新用户帐户
					try {
						server.createNewUser(user.name, user.password);
					} catch (Exception f) {
						sendDisconnect(0,
								strings.get(thisClass, "nonewaccount"));

						// 在我们断开客户的连接之前，给他一点时间来接收消息。
						try {
							sleep(1000);
						} catch (InterruptedException g) {
						}

						shutdown();
						return;
					}
					authenticated = true;
				}
			}

			if (!authenticated) {
				// “用户/密码”组合不匹配。
				sendDisconnect(0, strings.get(thisClass, "badpassword"));
				InetAddress tmpAddr = socket.getInetAddress();
				server.serverOutput(strings.get(thisClass, "failedlogin")
						+ " \"" + user.name + "\" "
						+ server.strings.get(thisClass, "fromhost") + " "
						+ tmpAddr.getHostName() + "\n");

				// 在我们断开客户的连接之前，给他一点时间来接收消息。
				try {
					sleep(1000);
				} catch (InterruptedException e) {
				}

				shutdown();
				return;
			}
		}

		synchronized (server.connections) {
			// 告诉我们的用户关于他们的情况。
			sendUserInfo(user);

			// 将聊天室列表发送给用户。
			sendRoomList();

			// 设置我们的聊天室指针。
			user.chatroom = server.mainRoom;

			// 将用户添加到聊天室。
			user.chatroom.addUser(this);

			// 告诉我们的用户他们在哪个房间。
			sendEnterRoom(user.id, user.chatroom);

			//我们需要告诉所有其他用户关于我们的信息，反之亦然。
			for (int count = 0; count < server.currentConnections; count++) {
				tmpConnection = (DushiServerConnection) server.connections
						.elementAt(count);

				// 告诉其他用户我们已经连接，并告诉他们我们的情况。
				tmpConnection.sendConnect(user.name);
				tmpConnection.sendUserInfo(user);
				tmpConnection.sendEnterRoom(user.id, user.chatroom);

				// 还可以向我们的用户介绍他们自己。
				sendUserInfo(tmpConnection.user);
				sendEnterRoom(tmpConnection.user.id,
						tmpConnection.user.chatroom);
			}

			// 将此用户添加到连接列表。
			server.connections.addElement(this);

			// 更新我们的统计数据。
			server.currentConnections = server.connections.size();
			server.totalConnections++;

			// 是否已超过连接数的峰值？
			if (server.currentConnections > server.peakConnections)
				server.peakConnections = server.currentConnections;
		}

		server.serverOutput(strings.get(thisClass, "newuser") + " " + user.name
				+ " " + strings.get(thisClass, "loggingonat") + " "
				+ dateFormatter.format(connectDate) + " "
				+ strings.get(thisClass, "from") + " "
				+ socket.getInetAddress().getHostName() + "\n");
		server.serverOutput(strings.get(thisClass, "thereare") + " "
				+ server.currentConnections + " "
				+ strings.get(thisClass, "usersconnected") + "\n");

		if (server.graphics) {
			// 更新服务器窗口。
			server.window.userList.add(user.name);
			server.window.disconnectAll.setEnabled(true);
			server.window.updateStats();
		}

		// 将突发消息发送给用户。
		sendServerText(strings.get(thisClass, "welcome") + " " + Dushi.VERSION
				+ ".\n\n");

		// 发送欢迎信息（如果有的话）。
		try {
			File welcomeFile = new File(DushiServer.welcomeFileName);
			BufferedReader in = new BufferedReader(new FileReader(welcomeFile));
			String inputLine = new String("");
			String input = new String("");

			while ((inputLine = in.readLine()) != null)
				input = input.concat(inputLine + "\n");
			in.close();

			// 输出。
			sendServerText(input + "\n");
		} catch (Exception a) {
		}

		String mess = "";
		if (server.currentConnections == 1)
			// 告知，他们是独有的用户。
			mess = strings.get(thisClass, "firstuser");

		else if (server.currentConnections > 2) {
			// 说明有多少用户在线，以及有多少聊天室（如果适用）。
			mess = strings.get(thisClass, "thereare") + " "
					+ (server.currentConnections - 1) + " "
					+ strings.get(thisClass, "otherusers");
			if (server.chatRooms.size() > 0)
				mess += ", " + server.chatRooms.size() + " "
						+ strings.get(thisClass, "chatrooms");
		} else {
			// 还有一个其他用户，可能在另一个聊天室。
			mess = strings.get(thisClass, "oneother");
			if (server.chatRooms.size() > 0)
				mess += " " + strings.get(thisClass, "inanother");
		}
		sendServerText(mess + "\n\n");

		// 在我们的聊天室捕获画布。
		catchUpWithCanvas(user.chatroom);

		// 是否有存储的消息正在等待此用户名？
		int numberMessages = 0;
		for (int count = 0; count < server.messages.size(); count++)
			if (((DushiMessage) server.messages.elementAt(count)).to
					.equals(user.name))
				numberMessages++;
		if (numberMessages > 0)
			sendServerMessage(strings.get(thisClass, "youhave") + " "
					+ numberMessages + " "
					+ strings.get(thisClass, "messageswaiting"));
		// 准备就绪
	}

	private void receiveDisconnect() throws IOException {
		// 用户正在断开与服务器的连接。将流清理干净。
		iStream.readInt();
		iStream.readUTF();
		iStream.readFully(new byte[iStream.available()]);

		// 断开连接。很显然，不要试图通知这个用户。
		server.disconnect(this, false);
	}

	protected void sendRoomList() {
		Vector tmpRooms = new Vector();

		tmpRooms.add(server.mainRoom);
		tmpRooms.addAll(server.chatRooms);

		super.sendRoomList(tmpRooms);
	}

	private void receiveRoomList() throws IOException {
		// 客户端正在请求可用聊天室的列表。从流中读取空“数”值。
		iStream.readShort();

		// 现在我们发送列表。
		sendRoomList();
	}

	private void sendInvite(int userId, String roomName) {
		super.sendInvite(userId, roomName, user.id);
	}

	private void receiveInvite() throws IOException {
		// 客户端希望邀请另一个用户进入聊天室。

		String roomName = "";
		DushiChatRoom inviteRoom;
		int inviteUserId = 0;
		DushiServerConnection inviteUser = null;

		iStream.readInt(); // user id
		roomName = iStream.readUTF();
		inviteUserId = iStream.readInt();

		// 查找请求的聊天室。
		inviteRoom = server.findChatRoom(roomName);

		if (inviteRoom == null) {
			// 我们猜想，根本没有这样的聊天室。
			sendServerMessage(strings.get(thisClass, "thechatroom") + " \""
					+ roomName + "\" " + strings.get(thisClass, "doesntexist"));
			return;
		}

		// 检查是否允许请求用户执行此操作。用户只能邀请其他人进入自己的房间。
		// 当然，这是有缺陷的，因为它只基于名称（而不是id），但目前它会这样做。
		if (!inviteRoom.creatorName.equals(user.name)
				&& (this != server.adminConnection)) {
			// 这不是此聊天室的所有者。叫他们走开。
			sendServerMessage(strings.get(thisClass, "noban"));
			return;
		}

		// 查找与用户id关联的客户端套接字。
		inviteUser = findConnection(inviteUserId);

		if (inviteUser == null) {
			// 嗯，找不到用户
			sendServerMessage(strings.get(thisClass, "nosuchuser"));
			return;
		}

		// 将用户的姓名添加到聊天室的禁止用户列表中。
		inviteRoom.inviteUser(inviteUser.user.id);

		// 向用户发送更新的聊天室列表。
		inviteUser.sendRoomList();

		// 向受邀用户发送邀请通知。
		inviteUser.sendInvite(user.id, roomName);
	}

	private void sendEnterRoom(int userId, DushiChatRoom room) {
		super.sendEnterRoom(userId, room.name, room.priv, "", false);
	}

	private void receiveEnterRoom() throws IOException {
		// 客户端正在请求进入/创建一个新的聊天室。
		// 房间的名称。
		String newRoomName = "";
		boolean priv = false;
		String password = "";
		DushiChatRoom newRoom;
		DushiServerConnection tmpConnection = null;

		iStream.readInt(); // 用户id
		newRoomName = iStream.readUTF();
		priv = iStream.readBoolean();
		password = iStream.readUTF();
		if (!iStream.readBoolean())
			password = server.passwordEncryptor.encryptPassword(password);

		// 查找请求的新聊天室。
		newRoom = server.findChatRoom(newRoomName);

		if (newRoom == null) {
			// 我们想根本没有这样的房间。我们将创造它。

			// 确保请求的名称是合法的。
			if (newRoomName.equals("")) {
				sendServerMessage(strings.get(thisClass, "needroomname"));
				return;
			}

			// 如果是私有的，则必须有密码
			if (priv && password.equals("")) {
				sendServerMessage(strings.get(thisClass, "needpassword"));
				return;
			}

			// 创建聊天室。
			newRoom = new DushiChatRoom(newRoomName, user.name, priv, password);
			try {
				newRoom.setLogging(server.logChats);
			} catch (IOException e) {
				server.serverOutput(strings.get(thisClass, "nolog") + " "
						+ newRoom.name + "\n");
			}

			// 将其添加到聊天室列表中。
			server.chatRooms.addElement(newRoom);

			synchronized (server.connections) {
				// 将更新后的聊天室列表发送给所有客户端。
				for (int count = 0; count < server.currentConnections; count++) {
					tmpConnection = (DushiServerConnection) server.connections
							.elementAt(count);
					tmpConnection.sendRoomList();
				}
			}
		} else {
			// 房间已存在。确保该用户被允许进入该房间
			//（即，他们没有被禁止，或其他什么）
			if (!server.isUserAllowed(newRoom, this, password))
				return;
		}

		// 离开我们当前的聊天室。
		if (user.chatroom != null)
			leaveChatRoom();

		// 加入聊天室。
		enterChatRoom(newRoom);

		// 保存它。
		user.chatroom = newRoom;
	}

	private void sendBootUser(int userId, String roomName) {
		super.sendBootUser(userId, roomName, user.id);
	}

	private void receiveBootUser() throws IOException {
		// 客户端希望从聊天室启动另一个用户。

		String roomName = "";
		DushiChatRoom bootRoom;
		int bootUserId = 0;
		DushiServerConnection bootUser = null;

		iStream.readInt(); // 用户id
		roomName = iStream.readUTF();
		bootUserId = iStream.readInt();

		// 查找请求的聊天室。
		bootRoom = server.findChatRoom(roomName);

		if (bootRoom == null) {
			// 我们猜想，根本没有这样的聊天室。
			sendServerMessage(strings.get(thisClass, "thechatroom") + " \""
					+ roomName + "\" " + strings.get(thisClass, "doesntexist"));
			return;
		}

		// 检查是否允许请求用户执行此操作。用户只能从自己的聊天室启动其他人。
		// 当然，这是有缺陷的，因为它只基于名称（而不是id），但目前它会这样做。
		if (!bootRoom.creatorName.equals(user.name)
				&& (this != server.adminConnection)) {
			// 这不是此聊天室的所有者。叫他们走开。
			sendServerMessage(strings.get(thisClass, "noboot"));
			return;
		}

		// 查找与用户id关联的客户端套接字。
		bootUser = findConnection(bootUserId);

		if (bootUser == null) {
			// 嗯，找不到用户。
			sendServerMessage(strings.get(thisClass, "nosuchuser"));
			return;
		}

		// 向用户发送一条消息，告知他们正在启动。
		bootUser.sendBootUser(user.id, roomName);

		// 如果这是主聊天室，请断开用户的连接。
		if (bootRoom == server.mainRoom) {
			server.disconnect(bootUser, true);
			return;
		}

		// 否则，请将该用户从聊天室中删除。
		else {
			bootUser.leaveChatRoom();

			// 将用户放回主聊天室。
			bootUser.enterChatRoom(server.mainRoom);

			bootUser.user.chatroom = server.mainRoom;
		}
	}

	/**
	 * 告诉客户他们被禁止进入聊天室。
	 */
	protected void sendBanUser(int userId, String roomName) {
		super.sendBanUser(userId, roomName, user.id);
	}

	private void receiveBanUser() throws IOException {
		// 客户端希望禁止其他用户进入聊天室。

		String roomName = "";
		DushiChatRoom banRoom;
		int banUserId = 0;
		DushiServerConnection banUser = null;

		iStream.readInt(); // 用户id
		roomName = iStream.readUTF();
		banUserId = iStream.readInt();

		// 查找请求的聊天室。
		banRoom = server.findChatRoom(roomName);

		if (banRoom == null) {
			// 我们猜想根本没有这样的i聊天室。
			sendServerMessage(strings.get(thisClass, "thechatroom") + " \""
					+ roomName + "\" " + strings.get(thisClass, "doesntexist"));
			return;
		}

		// 检查是否允许请求用户执行此操作。用户只能禁止其他人进入自己的聊天室。
		// 当然，这是有缺陷的，因为它只基于名称（而不是id），但目前它会这样做。
		if (!banRoom.creatorName.equals(user.name)
				&& (this != server.adminConnection)) {
			// 这不是此聊天室的所有者。叫他们走开。
			sendServerMessage(strings.get(thisClass, "noban"));
			return;
		}

		// 查找与用户id关联的客户端套接字。
		banUser = findConnection(banUserId);

		if (banUser == null) {
			// 嗯，找不到用户。
			sendServerMessage(strings.get(thisClass, "nosuchuser"));
			return;
		}

		// 将用户的姓名添加到聊天室的禁止用户列表中。
		banRoom.banUserName(banUser.user.name);

		// 我们不需要通知被禁止的用户；
		// 只有当他们试图进入被禁止进入的聊天室时，他们才会收到信息。
	}

	protected void sendAllowUser(int userId, String roomName) {
		super.sendAllowUser(userId, roomName, user.id);
	}

	private void receiveAllowUser() throws IOException {
		// 客户端希望允许被禁止的用户返回聊天室。

		String roomName = "";
		DushiChatRoom allowRoom;
		int allowUserId = 0;
		DushiServerConnection allowUser = null;

		iStream.readInt(); // 用户id
		roomName = iStream.readUTF();
		allowUserId = iStream.readInt();

		// 查找请求的聊天室。
		allowRoom = server.findChatRoom(roomName);

		if (allowRoom == null) {
			// 我们猜想没有这样的聊天室了。
			sendServerMessage(strings.get(thisClass, "thechatroom") + " \""
					+ roomName + "\" " + strings.get(thisClass, "doesntexist"));
			return;
		}

		// 检查是否允许请求用户执行此操作。用户只能授权他人使用自己拥有的房间。
		// 当然，这是有缺陷的，因为它只基于名称（而不是id），但目前它会这样做。
		if (!allowRoom.creatorName.equals(user.name)
				&& (this != server.adminConnection)) {
			// 这不是此聊天室的所有者。叫他们走开。
			sendServerMessage(strings.get(thisClass, "noallow"));
			return;
		}

		// 查找与用户id关联的客户端套接字。
		allowUser = findConnection(allowUserId);

		if (allowUser == null) {
			// 嗯，找不到用户。
			sendServerMessage(strings.get(thisClass, "nosuchuser"));
			return;
		}

		// 该用户是否在禁止用户列表中？
		for (int count = 0; count < allowRoom.bannedUserNames.size(); count++) {
			if (allowRoom.bannedUserNames.elementAt(count).equals(
					allowUser.user.name)) {
				// 从聊天室的禁止用户列表usersRe中i清除用户的名称。
				allowRoom.allowUserName(allowUser.user.name);

				// 通知用户现在允许他们进入房间。
				allowUser.sendAllowUser(user.id, roomName);

				break;
			}
		}
	}

	protected void sendActivity(int fromId, short activity) {
		// 清空收件人列表。
		super.sendActivity(fromId, activity, new int[0]);
	}

	private void receiveActivity() throws IOException {
		// 这将收到一个信号，表明我们的客户正忙于进行一些活动，如打字或绘图，
		// 并将其输出给适当的收件人。

		short activity = 0;
		int numRecipients = 0;
		DushiServerConnection toUser;

		// 读取发件人用户id（我们的）并将其丢弃。然后读取活动和收件人数量。
		iStream.readInt();
		activity = iStream.readShort();
		numRecipients = iStream.readInt();

		// 这是公开行动吗？如果是，请将其发送给当前聊天室中的所有人。
		if (numRecipients == 0) {
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					toUser.sendActivity(user.id, activity);
			}
		} else {
			// 每个预定收件人的循环
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());

				if (toUser != null)
					// 发送给此收件人
					toUser.sendActivity(user.id, activity);
			}
		}
	}

	protected void sendChatText(int fromId, boolean priv, String data) {
		// 清空收件人列表
		super.sendChatText(fromId, priv, data, new int[0]);
	}

	private void receiveChatText() throws IOException {
		// 这将接收一行聊天文本，并将其输出给相应的收件人。

		boolean priv = false;
		String data = "";
		int numRecipients = 0;
		DushiServerConnection toUser;

		// 读取发件人用户id（我们的）并将其丢弃。
		// 然后读取隐私值、颜色、数据和收件人数量。
		iStream.readInt();
		priv = iStream.readBoolean();
		iStream.readShort(); // 目前未使用的颜色。
		data = iStream.readUTF();
		numRecipients = iStream.readInt();

		// 这是公开行动吗？如果是，请将其发送给当前聊天室中的所有人。
		if (!priv) {
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					toUser.sendChatText(user.id, priv, data);
			}

			// 如果适用，请将其写入聊天日志文件。
			if (server.logChats)
				try {
					user.chatroom.log(user.name + "> " + data);
				} catch (IOException e) {
					server.serverOutput(strings.get(thisClass, "nowritelog")
							+ "\n");
					return;
				}
		} else {
			// 每个预定收件人的循环。
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());

				if (toUser != null)
					// 将其发送给此收件人。
					toUser.sendChatText(user.id, priv, data);
			}
		}
	}

	protected void sendLine(int fromId, short colour, short startX,
			short startY, short endX, short endY, short thick) {
		// 清空收件人列表
		super.sendLine(fromId, colour, startX, startY, endX, endY, thick,
				new int[0]);
	}

	private void receiveLine() throws IOException {
		// 我们的客户在屏幕上画了一条线。
		// 我们需要将其读入并发送给其他用户。

		short colour = 0;
		short startx = 0;
		short starty = 0;
		short endx = 0;
		short endy = 0;
		short thick = 0;
		int numRecipients = 0;
		DushiServerConnection toUser;

		// 读取发件人用户id（我们的）并将其丢弃。
		// 然后读取颜色、x/y坐标、厚度和收件人数量。
		iStream.readInt();
		colour = iStream.readShort();
		startx = iStream.readShort();
		starty = iStream.readShort();
		endx = iStream.readShort();
		endy = iStream.readShort();
		thick = iStream.readShort();
		numRecipients = iStream.readInt();

		// 这是公开行动吗？如果是，请将其发送给当前聊天室中的所有人。
		if (numRecipients == 0) {
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					toUser.sendLine(user.id, colour, startx, starty, endx,
							endy, thick);
			}

			// 将此行添加到画布向量中
			user.chatroom.canvasHistory.addElement(new CanvasCommand(
					DushiCommand.LINE, user.id, colour, startx, starty, endx,
					endy, (short) 0, (short) 0, (short) 0, thick, false, null));
		} else {
			// 现在为每个预定收件人循环
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());

				if (toUser != null)
					// 把它发给这个收件人
					toUser.sendLine(user.id, colour, startx, starty, endx,
							endy, thick);
			}
		}
	}

	protected void sendPoly(short kind, int fromId, short colour, short x,
			short y, short width, short height, short thick, boolean fill) {
		// 清空收件人列表
		super.sendPoly(kind, fromId, colour, x, y, width, height, thick, fill,
				new int[0]);
	}

	private void receivePoly(short kind) throws IOException {
		// 我们的客户在屏幕上画了一个多边形。
		// 我们需要将其读入并发送给其他用户。

		short colour = 0;
		short x = 0;
		short y = 0;
		short width = 0;
		short height = 0;
		short thick = 0;
		boolean fill = false;
		int numRecipients = 0;
		DushiServerConnection toUser;

		// 读取发件人用户id（我们的）并将其丢弃。
		// 然后读取颜色、x/y坐标、厚度和收件人数量。
		iStream.readInt();
		colour = iStream.readShort();
		x = iStream.readShort();
		y = iStream.readShort();
		width = iStream.readShort();
		height = iStream.readShort();
		thick = iStream.readShort();
		fill = iStream.readBoolean();
		numRecipients = iStream.readInt();

		// 这是公开行动吗？如果是，请将其发送给当前聊天室中的所有人。
		if (numRecipients == 0) {
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					toUser.sendPoly(kind, user.id, colour, x, y, width, height,
							thick, fill);
			}

			// 将此形状添加到画布vector中。 
			user.chatroom.canvasHistory.addElement(new CanvasCommand(kind,
					user.id, colour, x, y, width, height, (short) 0, (short) 0,
					(short) 0, thick, fill, null));
		} else {
			// 现在为每个预定收件人进行循环。
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());

				if (toUser != null)
					// 将其发送给此收件人。
					toUser.sendPoly(kind, user.id, colour, x, y, width, height,
							thick, fill);
			}
		}
	}

	protected void sendDrawText(int fromId, short colour, short x, short y,
			short type, short attribs, short size, String text) {
		// 清空收件人列表
		super.sendDrawText(fromId, colour, x, y, type, attribs, size, text,
				new int[0]);
	}

	private void receiveDrawText() throws IOException {
		// 我们的客户在屏幕上画了一些文字。
		// 我们需要将其读入并发送给其他用户。

		short colour = 0;
		short x = 0;
		short y = 0;
		short type = 0;
		short attribs = 0;
		short size = 0;
		String text = "";
		int numRecipients = 0;
		DushiServerConnection toUser;

		// 读取发件人用户id（我们的）并将其丢弃。
		// 然后读取颜色、x/y坐标、类型、大小和属性以及收件人数量。
		iStream.readInt();
		colour = iStream.readShort();
		x = iStream.readShort();
		y = iStream.readShort();
		type = iStream.readShort();
		attribs = iStream.readShort();
		size = iStream.readShort();
		text = iStream.readUTF();
		numRecipients = iStream.readInt();

		// 这是公开行动吗？如果是，请将其发送给当前聊天室中的所有人。
		if (numRecipients == 0) {
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					toUser.sendDrawText(user.id, colour, x, y, type, attribs,
							size, text);
			}

			// 将此文本添加到画布vector中。
			user.chatroom.canvasHistory.addElement(new CanvasCommand(
					DushiCommand.DRAWTEXT, user.id, colour, x, y, (short) 0,
					(short) 0, type, attribs, size, (short) 0, false, text));
		} else {
			// 现在为每个预定收件人进行循环。
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());

				if (toUser != null)
					// 发送给此收件人
					toUser.sendDrawText(user.id, colour, x, y, type, attribs,
							size, text);
			}
		}
	}

	private void receiveImage() throws IOException {
		// 我们的客户在屏幕上贴了一张图片。
		// 我们需要将其读入并发送给其他用户。

		String imageName = null;
		short xCoord = 0;
		short yCoord = 0;
		int totalSize = 0;
		int totalBlocks = 0;
		int blockSize = DushiImageTransfer.DEFAULT_BLOCKSIZE;
		byte[] data = null;
		int numRecipients = 0;
		Vector recipients = null;
		DushiServerConnection toUser = null;
		DushiImageTransfer transfer = null;

		// 读取发件人用户id（我们的）并将其丢弃。
		iStream.readInt();

		if (protocolVersion >= 2.2)
			// 读取图像名称
			imageName = iStream.readUTF();
		else
			// 制作一个
			imageName = (user.name + Math.random());

		// 读取x和y坐标。
		xCoord = iStream.readShort();
		yCoord = iStream.readShort();

		if (protocolVersion >= 2.2) {
			// 图像数据以单独的IMAGEBLOCK命令发送。
			// 在这里，我们刚刚阅读了我们将收到的邮件数量。
			totalSize = iStream.readInt();
			totalBlocks = iStream.readInt();
			blockSize = iStream.readInt();
		} else {
			// 读取实际图像数据。
			totalSize = iStream.readInt();
			totalBlocks = ((totalSize / blockSize) + (((totalSize % blockSize) > 0) ? 1
					: 0));
			data = new byte[totalSize];
			int read = 0;
			while (read < totalSize) {
				int tmp = iStream.read(data, read, (totalSize - read));
				if (tmp < 1)
					throw (new EOFException());
				read += tmp;
			}
		}

		numRecipients = iStream.readInt();
		recipients = new Vector();

		// 这是公开动作吗？如果是，请将其发送给当前聊天室中的所有人。
		if (numRecipients == 0) {
			// 针对聊天室中每个人的进行循环（除了我们）。
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					recipients.add(toUser);
			}
		} else {
			// 针对每个预定收件人的进行循环
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());
				if (toUser != null)
					// 将其发送给此收件人。
					recipients.add(toUser);
			}
		}

		transfer = new DushiImageTransfer(user.id, imageName, xCoord, yCoord,
				totalSize, totalBlocks, blockSize, recipients, new int[0],
				null, null);
		if (numRecipients != 0)
			transfer.priv = true;

		if (protocolVersion >= 2.2) {
			// 把它放在我们的图像传输列表中，等待所有数据到达。
			imageTransfers.put(imageName, transfer);
		} else {
			//我们已经掌握了所有的数据。
			transfer.putImageAll(data);

			if (!transfer.priv)
				// 将此图片添加到我们的画布历史vector中。
				user.chatroom.canvasHistory.addElement(new CanvasCommand(
						DushiCommand.IMAGE, user.id, xCoord, yCoord, imageName,
						data));
		}
	}

	protected void sendClearCanv(int fromId) {
		// 清空收件人列表
		super.sendClearCanv(fromId, new int[0]);
	}

	private void receiveClearCanv() throws IOException {
		// 这将收到一个信号，表明我们的客户希望清除画布。

		int numRecipients = 0;
		DushiServerConnection toUser;

		//读取发件人用户id（我们的）并将其丢弃。然后读取收件人的数量。
		iStream.readInt();
		numRecipients = iStream.readInt();

		// 这是公共动作吗？如果是，请将其发送给当前聊天室中的所有人。
		if (numRecipients == 0) {
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					toUser.sendClearCanv(user.id);
			}

			// 清空我们的画布vector。
			user.chatroom.canvasHistory.removeAllElements();
		} else {
			// 针对每个预定收件人的进行循环
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());

				if (toUser != null)
					// 发送给此收件人
					toUser.sendClearCanv(user.id);
			}
		}
	}

	protected void sendPageUser(int fromId) {
		// 清空收件人列表
		super.sendPageUser(fromId, new int[0]);
	}

	private void receivePageUser() throws IOException {
		// 这将接收到一个信号，即我们的客户希望寻呼一个或多个其他用户。

		int numRecipients = 0;
		DushiServerConnection toUser;

		// 读取发件人用户id（我们的）并将其丢弃。然后读取收件人的数量。
		iStream.readInt();
		numRecipients = iStream.readInt();

		// 这是公开行动吗？如果是，请将其发送给当前聊天室中的所有人。
		if (numRecipients == 0) {
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					toUser.sendPageUser(user.id);
			}
		} else {
			// 针对每个预定收件人的进行循环。
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());

				if (toUser != null)
					// 发送给此收件人
					toUser.sendPageUser(user.id);
			}
		}
	}

	protected void sendInstantMess(int fromId, String message) {
		super.sendInstantMess(fromId, user.id, message);
	}

	private void receiveInstantMess() throws IOException {
		int whoFor = 0;
		String message = "";

		// 我们的客户希望向另一个用户发送即时消息。
		// 首先，读取我们自己的用户ID并将其丢弃。
		// 然后，读取收件人的姓名和消息文本。
		iStream.readInt();
		whoFor = iStream.readInt();
		message = iStream.readUTF();

		synchronized (server.connections) {
			// 找到收件人并发送消息。
			for (int count = 0; count < server.currentConnections; count++) {
				DushiServerConnection toUser = (DushiServerConnection) server.connections
						.elementAt(count);
				if (toUser.user.id == whoFor) {
					toUser.sendInstantMess(user.id, message);
					break;
				}
			}
		}
	}

	private void receiveLeaveMess() throws IOException {
		String whofor = "";
		String message = "";

		// 我们的客户想给另一个用户留言。
		// 首先，读取我们自己的用户ID并将其丢弃。
		// 然后，读取收件人的姓名和消息文本。
		iStream.readInt();
		whofor = iStream.readUTF();
		message = iStream.readUTF();

		// 将其添加到消息列表中
		server.messages
				.addElement(new DushiMessage(whofor, user.name, message));
	}

	private void receiveReadMess() throws IOException {
		// 客户正在请求他的消息

		// 读取发件人用户id（我们的）并将其丢弃。
		iStream.readInt();

		// 现在发送用户存储的所有消息。
		sendStoredMess();
	}

	protected void sendStoredMess() {
		// 这将发送为该用户存储的所有消息。

		DushiMessage message;
		Vector messages = new Vector();

		Enumeration e = server.messages.elements();
		while (e.hasMoreElements()) {
			message = (DushiMessage) e.nextElement();

			if (message.to.equals(user.name)) {
				messages.add(message);
				server.messages.remove(message);
			}
		}

		super.sendStoredMess(messages);
	}

	protected void sendError(int fromId, short errorCode) {
		// 清空收件人列表。
		super.sendError(fromId, errorCode, new int[0]);
	}

	private void receiveError() throws IOException {
		// 这将从我们的客户端收到一条错误消息，
		// 并将其转发给相应的用户。

		int numRecipients = 0;
		short errorCode = 0;
		DushiServerConnection toUser;

		// 读取发件人用户id（我们的）并将其丢弃。
		iStream.readInt();
		errorCode = iStream.readShort();
		numRecipients = iStream.readInt();

		// 这是公共动作吗？如果是，请将其发送给当前聊天室中的所有人。
		if (numRecipients == 0) {
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					toUser.sendError(user.id, errorCode);
			}
		} else {
			// 针对每个预定收件人的做循环。
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());

				if (toUser != null)
					// 将其发送给此收件人。
					toUser.sendError(user.id, errorCode);
			}
		}
	}

	private void receiveSendEmail() throws IOException {
		// 这将收到我们客户的“发送电子邮件”请求，并启动DushiEmail线程进行发送。

		String mxHost = null;
		String fromAddress = null;
		String toAddress = null;
		String subject = null;
		String text = null;
		String hostName = null;
		String fromName = null;
		DushiEmail email = null;

		// 阅读发件人用户id（我们的）并丢弃它。
		// 然后阅读收件人：地址、主题和消息文本。
		iStream.readInt();
		mxHost = iStream.readUTF();
		fromAddress = iStream.readUTF();
		toAddress = iStream.readUTF();
		subject = iStream.readUTF();
		text = iStream.readUTF();

		// 现在尝试发送电子邮件。
		try {
			hostName = InetAddress.getLocalHost().getCanonicalHostName();
		} catch (UnknownHostException e) {
			String tmp = strings.get(thisClass, "errorsending") + "\n"
					+ toAddress + ": \"" + subject + "\":\n" + e;
			sendServerMessage(tmp);
			server.serverOutput(tmp + "\n");
		}

		if ((fromAddress == null)
				|| ((fromAddress != null) && (fromAddress.length() <= 0)))
			fromAddress = "Dushi_user@" + hostName;
		fromName = user.name;

		// 创建电子邮件
		email = new DushiEmail(this, fromAddress, toAddress, subject,
				"Sent by Dushi Chat user " + user.name + "@" + hostName
						+ "\n\n" + text);

		// 如果为我们指定了MX主机，请将其添加到电子邮件中
		if ((mxHost != null) && (mxHost.length() > 0))
			email.mxHost = mxHost;
		if (fromName != null)
			email.fromName = fromName;

		// 发送电子邮件
		email.start();
	}

	protected void sendFile(int fromId, String fileName, long totalSize,
			int totalBlocks, int blockSize, int expireMins) {
		// 清空收件人列表
		super.sendFile(fromId, fileName, totalSize, totalBlocks, blockSize,
				expireMins, new int[0]);
	}

	private void receiveFile() throws IOException {
		// 这将从我们的客户端接收“文件发送”请求，实例化文件传输线程，
		// 并将请求转发给列出的用户。

		String fileName = null;
		long totalSize = 0;
		int totalBlocks = 0;
		int blockSize = 0;
		int expireMins = 0;
		int numRecipients = 0;
		Vector toUsers = new Vector();
		DushiServerConnection toUser;

		// 读取发件人用户id（我们的）并将其丢弃。
		// 然后读取文件名、总块数、块大小、过期分钟数和收件人数。
		iStream.readInt();
		fileName = iStream.readUTF();
		totalSize = iStream.readLong();
		totalBlocks = iStream.readInt();
		blockSize = iStream.readInt();
		expireMins = iStream.readInt();
		numRecipients = iStream.readInt();

		// 这是公开行动吗？如果是，我们会将其发送给当前聊天室中的每个人。
		if (numRecipients == 0) {
			for (int count = 0; count < user.chatroom.users.size(); count++) {
				toUser = (DushiServerConnection) user.chatroom.users
						.elementAt(count);
				if (toUser != this)
					toUsers.add(toUser);
			}
		} else {
			// 对每个预定收件人的做循环。
			for (int count = 0; count < numRecipients; count++) {
				toUser = findConnection(iStream.readInt());

				if (toUser != null)
					toUsers.add(toUser);
			}
		}

		if (toUsers.size() <= 0)
			// 嗯？发送给任何人。别为这个烦恼。
			return;

		if (expireMins > 30)
			// 在此处设置任意最大值
			expireMins = 30;

		// 创建新的文件传输线程。
		fileTransfers.put(fileName, new DushiServerFileTransfer(this, fileName,
				totalSize, totalBlocks, blockSize, expireMins, toUsers));

		// 现在将请求转发给指定的用户。
		for (int count = 0; count < toUsers.size(); count++) {
			toUser = (DushiServerConnection) toUsers.elementAt(count);
			toUser.sendFile(user.id, fileName, totalSize, totalBlocks,
					blockSize, expireMins);
		}
	}

	private void sendFileAccept(int fromId, String fileName, boolean accept) {
		super.sendFileAccept(fromId, user.id, fileName, accept);
	}

	private void receiveFileAccept() throws IOException {
		// 这将从我们的客户端接收“文件接受”请求（对另一个用户的“文件发送”文件
		// 传输请求的响应），如果答案为“true”，则将我们的客户端套接字添加到传
		// 输的参与者列表中。

		int toUserId = 0;
		String fileName = null;
		boolean accept = false;
		DushiServerConnection toUser;
		DushiServerFileTransfer transfer = null;

		// 读取发件人用户id（我们的）并将其丢弃，然后读取收件人id、
		// 文件名以及我们是否接受该文件。
		iStream.readInt();
		toUserId = iStream.readInt();
		fileName = iStream.readUTF();
		accept = iStream.readBoolean();

		// 现在查找收件人
		toUser = findConnection(toUserId);

		if (toUser == null) {
			// 嗯，找不到用户
			sendServerMessage(strings.get(thisClass, "nosuchuser"));
			return;
		}

		if (accept) {
			// 我们正在接受文件，请尝试在我们接受的用户列表中查找文件传输。
			transfer = (DushiServerFileTransfer) toUser.fileTransfers
					.get(fileName);
			if (transfer != null)
				transfer.addUser(this);
		}

		// 转发验收通知
		toUser.sendFileAccept(user.id, fileName, accept);
	}

	private void receiveFileBlock() throws IOException {
		// 我们的客户正在发送一个文件块。
		// 找到由“发送文件”请求启动的文件传输线程，读取块，
		// 并将其添加到文件传输中。

		String fileName = null;
		int blockNumber = 0;
		int dataLength = 0;
		byte[] data = null;
		DushiServerFileTransfer transfer = null;

		// 读取发送方用户id（我们的）并丢弃它。
		// 然后读取文件名、块号、块大小和数据。
		iStream.readInt();
		fileName = iStream.readUTF();
		blockNumber = iStream.readInt();
		dataLength = iStream.readInt();

		data = new byte[dataLength];
		int read = 0;
		while (read < dataLength) {
			int tmp = iStream.read(data, read, (dataLength - read));
			if (tmp < 1)
				throw (new EOFException());
			read += tmp;
		}

		transfer = (DushiServerFileTransfer) fileTransfers.get(fileName);
		if (transfer != null)
			transfer.putFileBlock(blockNumber, data);
	}

	private void receiveFileCancel() throws IOException {
		// 我们的客户告诉我们停止参与文件传输。
		// 我们做什么取决于我们是发送还是接收文件。

		String fileName = null;
		DushiServerFileTransfer transfer = null;

		// 读取发件人用户id（我们的）并丢弃它。
		// 然后读取文件名。
		iStream.readInt();
		fileName = iStream.readUTF();

		// 看看这是否是我们的文件传输
		transfer = (DushiServerFileTransfer) fileTransfers.get(fileName);
		if (transfer != null)
			transfer.cancel();
		else {
			synchronized (server.connections) {
				// 不，一定是别人转来的
				Enumeration e = server.connections.elements();

				while (e.hasMoreElements()) {
					transfer = (DushiServerFileTransfer) e.nextElement();

					if (transfer.fileName.equals(fileName)
							&& transfer.removeUser(this))
						break;
				}
			}
		}
	}

	private void receiveImageBlock() throws IOException {
		// 我们的客户正在发送一个图像块。
		// 找到由“图像”请求启动的图像传输线程，读取块，并将其添加到传输中。

		String imageName = null;
		int blockNumber = 0;
		int dataLength = 0;
		byte[] data = null;
		DushiImageTransfer transfer = null;

		// 读取发件人用户id（我们的）并丢弃它。
		// 然后读取图像名称、块号、块大小和数据。
		iStream.readInt();
		imageName = iStream.readUTF();
		blockNumber = iStream.readInt();
		dataLength = iStream.readInt();

		data = new byte[dataLength];
		int read = 0;
		while (read < dataLength) {
			int tmp = iStream.read(data, read, (dataLength - read));
			if (tmp < 1)
				throw (new EOFException());
			read += tmp;
		}

		transfer = (DushiImageTransfer) imageTransfers.get(imageName);
		if (transfer != null) {
			transfer.putImageBlock(blockNumber, data);

			if (transfer.gotAll) {
				if (!transfer.priv)
					// 将此图片添加到我们的画布历史vector中
					user.chatroom.canvasHistory.addElement(new CanvasCommand(
							DushiCommand.IMAGE, transfer.fromId,
							transfer.xCoord, transfer.yCoord,
							transfer.imageName, transfer.getImageAll()));
				imageTransfers.remove(transfer.imageName);
			}
		}
	}

	protected void lostConnection() {
		server.disconnect(this, false);
	}

	class CanvasCommand {
		// 这样我们就可以保持画布命令的vector来跟踪所有绘图。

		short command;
		int who;
		short x1, y1, x2, y2;
		short colour, type, attribs, size, thick;
		boolean fill;
		String text;
		String imageName;
		byte[] imageData;

		/**
		 * 除了图像以外的所有东西。
		 */
		public CanvasCommand(short cmd, int w, short c, short _x1, short _y1,
				short _x2, short _y2, short ty, short a, short s, short th,
				boolean f, String txt) {
			command = cmd;
			who = w;
			colour = c;
			x1 = _x1;
			y1 = _y1;
			x2 = _x2;
			y2 = _y2;
			type = ty;
			attribs = a;
			size = s;
			thick = th;
			fill = f;
			text = txt;
		}

		/**
		 * 对于图像。
		 */
		public CanvasCommand(short cmd, int w, short _x1, short _y1, String n,
				byte[] d) {
			command = cmd;
			who = w;
			x1 = _x1;
			y1 = _y1;
			imageName = n;
			imageData = d;
		}
	}
}
